/***************************************************************************
 *   fadprogress -- gui to display progress of FAD (TM) calculations       *
 *   Copyright (C) 2005 by Eric Smith                                      *
 *   Contact information:  MerePeer at www.free-dc.org                     *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   The GPL can usually be found at http://www.gnu.org/licenses/gpl.txt   *
 ***************************************************************************/

/*
Version history:
    v1.00	- creation from FADProgress

*/
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;

public class FADProgress extends JPanel 
	implements ActionListener, ItemListener
{
	final static String sVERSION = "1.00";
	final static String sPACKET_VERSION = "D04";
	final static String sPACKET_DELIM = "\t";
	final static String sPACKET_TYPE_FAD = "F";
	final static String sPACKET_TYPE_PROBLEM = "A";
	final static String sSTATS_PAGE = "http://stats.find-a-drug.com/livequery1.html";

	// -- Common packet
	private final int PKTVERSION = 0;
	private final int PKTPROGRESSDNAME = 1; // name of system sending progress
	private final int PKTNODENAME = 2; // name of fad node
	private final int PKTTYPE = 3;
	// -- FAD specific packet
	private final int PKTFADTARGET = 4;
	private final int PKTFADJOBNAME = 5;
	private final int PKTFADSTART = 6;
	private final int PKTFADLATEST = 7;
	private final int PKTFADPROGRESS = 9;
	private final int PKTFADHITS = 11;
	private final int PKTFADRATING = 12;

	final int iROW_HEIGHT = 19;
	final int iSTALE_UPDATE_DELAY = 180*1000; // if node update not received for this long, show warning color
	final int iSTALE_STATE_DELAY = 10*60*1000; // if state hasn't chgd for this long, show warning color
	final int iANCIENT_STATE_DELAY = 20*60*1000; // if state hasn't chgd for this long, show red bgnd
	final int iSIX_HOURS = 6*60*60*1000;

	public static boolean cfgDEBUG = false;
	public static int cfgPORT = 9031; // listen on this port for incoming progress node udp packets
	public static int cfgTIMER = 10*1000; // refresh window this # seconds
	public static int cfgHTML_TIMER = 10*1000; // refresh html output this # seconds (if checked)
	public static int cfgHTML_REFRESH = 0; // 0 means dont meta refresh.  If set, this value will be used.
	public static Boolean cfgHTML_ENABLED = new Boolean(false); // should html output be started when the app starts?
	public static String cfgUSERNAME = null; // fad nickname - future

	public static JFrame jframe;

	protected NumberFormat nFormat = NumberFormat.getNumberInstance(Locale.getDefault());
	protected Font fontRegular = new Font("Arial", Font.PLAIN, 12);
	protected Color cAlternatingBgnd = new Color(Integer.parseInt("E8F0F0", 16));
	protected JLabel jlPoints;
	protected JLabel jlRank;
	protected int iPoints = 0;
	protected FADTableModel fadtm;
	protected JCheckBoxMenuItem jcbHTML;
	protected javax.swing.Timer timerHTML;
	//	protected JLabel jlStatus;

	//----------------------------------------------------------------------
	public FADProgress() // class constructor
	{
		// create ourself as a border layout with 5hz and 5vt spacing
		super(new BorderLayout(5,5));

		// ---------- MAIN PANEL SOUTH
// 		JPanel jpanelSouth = new JPanel(new FlowLayout(FlowLayout.LEFT));
// 		jlStatus = new JLabel("");
// 		jlStatus.setFont(fontRegular);
// 		jlStatus.setPreferredSize(new Dimension(800, 20));
// 		jlStatus.setBorder(BorderFactory.createLoweredBevelBorder());
// 		jpanelSouth.add(jlStatus);
// 		this.add(jpanelSouth, BorderLayout.SOUTH);

		// Other init
		nFormat.setMaximumFractionDigits(1);

		// ---------- MAIN PANEL NORTH
 		JPanel jpanelNorth = new JPanel(new FlowLayout(FlowLayout.LEFT));
 		JLabel jl = new JLabel("");
 		jl.setFont(fontRegular);
 		jpanelNorth.add(jl);
// 		jlPoints = new JLabel("", JLabel.RIGHT);
//  		jlPoints.setFont(fontRegular);
//  		jlPoints.setBorder(BorderFactory.createLoweredBevelBorder());
//  		jlPoints.setPreferredSize(new Dimension(90, 20));
// 		jpanelNorth.add(jlPoints);

// 		jlRank = new JLabel("", JLabel.CENTER); // use this for rank as of
//  		jlRank.setFont(fontRegular);
//  		jlRank.setBorder(BorderFactory.createLoweredBevelBorder());
// 		jpanelNorth.add(jlRank);

		this.add(jpanelNorth, BorderLayout.NORTH);

		// --------- MAIN PANEL CENTER
		// create the center panel as a border layout with 5hz and 5vt spacing
		JPanel jpanelCenter = new JPanel(new BorderLayout(5,5));

		// create the table
		fadtm = new FADTableModel();
		JTable jtable = new JTable(fadtm)
		{
			protected JTableHeader createDefaultTableHeader()
			{
				return new JTableHeader(columnModel)
				{
					public String getToolTipText(MouseEvent e)
					{
						java.awt.Point p = e.getPoint();
						int iColIndex = columnModel.getColumnIndexAtX(p.x);
						int iModelIndex = columnModel.getColumn(iColIndex).getModelIndex();
						return fadtm.getColumnTip(iModelIndex);
					}
				};
			};
		};
		jtable.setPreferredScrollableViewportSize(new Dimension(935, 170));
		jtable.setGridColor(this.getBackground());
		jtable.setBackground(new Color(Integer.parseInt("FAFAFA", 16)));
		jtable.setFont(fontRegular);
		jtable.setRowHeight(iROW_HEIGHT);

		JTableHeader jheader = jtable.getTableHeader();
		jheader.setFont(fontRegular);
		Dimension d = jheader.getPreferredSize();
		d.height += 7; // make header row height higher
		jheader.setPreferredSize(d);
		jheader.addMouseListener(new FADHeaderMouseHandler());
//		jtable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

		FADProgressRenderer pRenderer = new FADProgressRenderer(true);
		FADNumberRenderer nRenderer = new FADNumberRenderer(); // integer
		FADNumberRenderer fRenderer = new FADNumberRenderer(true); // float
		FADHeaderRenderer hRenderer = new FADHeaderRenderer();
		FADTextRenderer tRenderer = new FADTextRenderer();
		FADDateRenderer dRenderer = new FADDateRenderer();
		FADDateRenderer dsRenderer = new FADDateRenderer(true); // show stale
		FADStateRenderer sRenderer = new FADStateRenderer();

		Enumeration enum = jtable.getColumnModel().getColumns();
		TableColumn tc = null;
		while (enum.hasMoreElements())
		{
			tc = ((TableColumn)enum.nextElement());
			tc.setHeaderRenderer(hRenderer);
			int iModel = tc.getModelIndex();
			tc.setPreferredWidth(fadtm.aiWidths[iModel]);
			switch (iModel)
			{
				// progress columns get progress renderer
				case FADCruncher.JOBPROGRESS:
					tc.setCellRenderer(pRenderer);
					break;
				// number columns get number renderer
				case FADCruncher.HITS:
				case FADCruncher.RATING:
					tc.setCellRenderer(nRenderer);
					break;
				// date renderer
				case FADCruncher.JOBSTART:
					tc.setCellRenderer(dRenderer);
					break;
				// date renderer showing stale colors
				case FADCruncher.ACTIVITY:
				case FADCruncher.UPDATED:
					tc.setCellRenderer(dsRenderer);
					break;
				// default is text renderer
				default:
					tc.setCellRenderer(tRenderer);
					break;
			}
		}

		// make a scroll pane with a nice border
		JScrollPane scrollPane = new JScrollPane(jtable);
		scrollPane.setBorder(BorderFactory.createLoweredBevelBorder());
		scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );

		// put the scroll pane in the middle of the center panel
		jpanelCenter.add(scrollPane, BorderLayout.CENTER);
		// now put empty labels around the center panel to beef up the margins so it isn't against the frame
		jpanelCenter.add(new JLabel(), BorderLayout.NORTH);
		jpanelCenter.add(new JLabel(), BorderLayout.SOUTH);
		jpanelCenter.add(new JLabel(), BorderLayout.EAST);
		jpanelCenter.add(new JLabel(), BorderLayout.WEST);

		this.add(jpanelCenter, BorderLayout.CENTER);
		this.setOpaque(true); // must be opaque

		// ++++++++++++++++++
		// this thread listens for incoming packets
		try
		{
			new FADListenThread("FADListener").start();
		} catch (IOException ex) {
			ex.printStackTrace();
			System.exit(-1);
		}

		// ++++++++++++++++++
		// this task updates the table
		javax.swing.Timer t = new javax.swing.Timer(cfgTIMER, new ActionListener()
			{
				public void actionPerformed(ActionEvent evt)
				{
					fadtm.fireTableDataChanged();
				}
			}
		);
		t.start();

		// +++++++++++++++++++
		// this task outputs html if the menu item has been checked
		timerHTML = new javax.swing.Timer(cfgHTML_TIMER, new ActionListener()
			{
				public void actionPerformed(ActionEvent evt)
				{
					//					System.out.println("in timer" + new Date());
					outputHTML();
				}
			}
      );
		timerHTML.setInitialDelay(0); // fire immediately when first started
		// dont start it -- that will be done by menu checkbox
	}

	//----------------------------------------------------------------------
	private static void createAndShowGUI()
	{
		// do all frame type stuff at the static level
		JMenuItem jmi;

		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			JFrame.setDefaultLookAndFeelDecorated(true);
		} catch (Exception e) { }

		jframe = new JFrame("FAD Progress");
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		FADProgress newContentPane = new FADProgress();

		JMenuBar jmb = new JMenuBar();
		JMenu jm = new JMenu("File");
		jmb.add(jm);
		(jmi = new JMenuItem("Exit")).addActionListener(newContentPane);
		jm.add(jmi);

		jm = new JMenu("Options");
		jmb.add(jm);
		(newContentPane.jcbHTML = new JCheckBoxMenuItem("Generate HTML File")).addItemListener(newContentPane);
		jm.add(newContentPane.jcbHTML);
		if (cfgHTML_ENABLED.booleanValue()) newContentPane.jcbHTML.doClick(); // turn it on if ini setting so desired
//		(jmi = new JMenuItem("Request Forwarding...")).addActionListener(newContentPane);
	//	jm.add(jmi);

		jm = new JMenu("Help");
		jmb.add(jm);
		(jmi = new JMenuItem("About...")).addActionListener(newContentPane);
		jm.add(jmi);

		jframe.setJMenuBar(jmb);
		JScrollPane jsp = new JScrollPane(newContentPane);
		jsp.setPreferredSize(new Dimension(955, 260));
		jframe.setContentPane(jsp);
		if (cfgUSERNAME != null) jframe.setTitle("FAD Progress - " + cfgUSERNAME);
		//jframe.setContentPane(newContentPane);
		jframe.pack();
		jframe.setVisible(true);
  	}

	//----------------------------------------------------------------------
	public static void main(String[] args)
	{
		int i = 0;

		// output version to console
		System.out.println("FADProgress v" + sVERSION);
		
		if ((args.length > 0) && (args[0].equals("-debug")))
		{
			System.out.println("Debug on.");
			cfgDEBUG = true;
		}

		// if there is an ini file, grab settings first from there
		readIniFile();

		// if there are command line args, override ini settings
		while ((i < args.length) && (args[i].startsWith("-")))
		{
			String sArg = args[i++];

			if (sArg.equals("-debug"))
			{
				System.out.println("Debug on.");
				cfgDEBUG = true;
			} else if (sArg.equals("-port")) {
				if (i < args.length)
				{
					String sPort = args[i++];
					try
					{
						cfgPORT = Integer.parseInt(sPort);
					} catch (Exception ex) {
						reportError(true, null, "Unable to convert -port argument (" + sPort + ") into an integer.");
					}
					if ((cfgPORT < 1024) || (cfgPORT > 32767))
					{
						reportError(true, null, "Please specify a port between 1024 and 32767, exclusive.");
					}
					if (cfgDEBUG) System.out.println("Using port: " + cfgPORT);
				} else {
					reportError(true, null, "-port requires a port number arg, i.e. -port 5030");
				}
			} else if (sArg.equals("-timer")) {
				if (i < args.length)
				{
					String sTimer = args[i++];
					try
					{
						cfgTIMER = Integer.parseInt(sTimer) * 1000;
					} catch (Exception ex) {
						reportError(true, null, "Unable to convert -timer argument (" + sTimer + ") into an integer.");
					}
					if ((cfgTIMER < 1000) || (cfgTIMER > 600000))
					{
						reportError(true, null, "Please specify a timer value between 1 and 600");
					}
					if (cfgDEBUG) System.out.println("Using timer: " + cfgTIMER / 1000);
				} else {
					reportError(true, null, "-timer requires a timer delay arg, i.e. -timer 6");
				}
			} // endif possible args
		} // end while
		// check for invalid args, or user wants help on args
		if (i != args.length) // should be same if it worked
		{
			System.err.println("Usage:  java -cp {dir where .class file is} FADProgress [-port aport] [-timer nseconds]");
			System.exit(0);
		}

		// Schedule a job for the event-dispatching thread to do all the work
		javax.swing.SwingUtilities.invokeLater(new Runnable()
			{ public void run() { createAndShowGUI(); }});
	}

	//----------------------------------------------------------------------
	public static void readIniFile()
	{
		BufferedReader br = null;
		try
		{
			br = new BufferedReader(new FileReader(new File("FADProgress.ini")));
			String sLine;
			String sOption;
			String sValue;
			int iIndex;
			int iCount = 0;
			while ((sLine = br.readLine()) != null)
			{
				if (cfgDEBUG) System.out.println(iCount++ + ": " + sLine);
				if (sLine.length() < 3) break; // must have a few chars in it
				iIndex = sLine.indexOf("=");
				if (iIndex < 1) break; // must have equals sign in it
				sOption = sLine.substring(0, iIndex).toLowerCase();
				sValue = sLine.substring(iIndex + 1);

				if (sOption.equals("username")) {
					cfgUSERNAME = sValue;
					if (cfgDEBUG) System.out.println("username is: " + cfgUSERNAME);

				} else if (sOption.equals("html_timer")) {
					try
					{
						cfgHTML_TIMER = 1000 * Math.max(10, Integer.parseInt(sValue));
					} catch (Exception ex) {
						JOptionPane.showMessageDialog(null, "Unable to convert your FADProgress.ini 'html_timer' value into a number: " + sValue, "", JOptionPane.WARNING_MESSAGE);
					}
					if (cfgDEBUG) System.out.println("cfgHTML_TIMER is: " + cfgHTML_TIMER);

				} else if (sOption.equals("html_refresh")) {
					try
					{
						cfgHTML_REFRESH = Math.max(10, Integer.parseInt(sValue));
					} catch (Exception ex) {
						JOptionPane.showMessageDialog(null, "Unable to convert your FADProgress.ini 'html_refresh' value into a number: " + sValue, "", JOptionPane.WARNING_MESSAGE);
					}
					if (cfgDEBUG) System.out.println("cfgHTML_REFRESH is: " + cfgHTML_REFRESH);

				} else if (sOption.equals("html_enabled")) {
					cfgHTML_ENABLED = new Boolean(sValue);
					if (cfgDEBUG) System.out.println("cfgHTML_ENABLED is: " + cfgHTML_ENABLED);					
				}
			}
			br.close();
		} catch (FileNotFoundException ex) {
		// do nothing -- its ok not to have an ini file
		} catch (Exception ex) {
			if (br != null) try { br.close(); } catch (Exception ex2) { };
			JOptionPane.showMessageDialog(null, "Problem accessing FADProgress.ini: " + ex, "", JOptionPane.WARNING_MESSAGE);
		}
	}

	// BEGIN NON STATIC METHODS
	//**********************************************************************
	public void addPoints(Integer iNewPoints)
	{
		iPoints += iNewPoints.intValue();
		jlPoints.setText(nFormat.format(iPoints));
	}

	//----------------------------------------------------------------------
	public void updatePoints()
	{
		int iReady = 0;
/*		for (int i=0; i < fadtm.getRowCount(); i++)
		{
			Object oMaybeNull = fadtm.getCruncherAt(i).getColumn(FADCruncher.RESULTSTOUPLOAD); // might be null
			if (oMaybeNull != null) iReady += ((Integer) oMaybeNull).intValue();
		}
		jlPoints.setText(nFormat.format(iReady));
		*/
	}

	//----------------------------------------------------------------------
	public void actionPerformed(ActionEvent e)
	{
		String sAction = e.getActionCommand();
		if ("About...".equals(sAction))
		{
			JOptionPane.showMessageDialog(this, "FADProgress v" + sVERSION +
					"\nUse at your own risk." +
					"\nFree software under the terms of the GNU General Public License." +
					"\nSee source code for copyright and license information.");
		} else if ("Exit".equals(sAction)) {
			System.exit(0);
		} else if ("Request Forwarding...".equals(sAction)) {
			//retrieveFADProjects();
			//		requestForwarding();
		}

	}

	//----------------------------------------------------------------------
	public void itemStateChanged(ItemEvent e)
	{
		Object source = e.getItemSelectable();
		if (source == jcbHTML)
		{
			if (e.getStateChange() == ItemEvent.DESELECTED)
			{
				timerHTML.stop();
			}
			else
			{
				timerHTML.start();
			}
		}
	}

	//----------------------------------------------------------------------
	public void outputHTML()
	{
		try
		{
			FileWriter fw = new FileWriter("FADProgress.html");
			fw.write("<html>\n");
			fw.write("<head>\n");
			if (cfgHTML_REFRESH > 0) fw.write("<meta http-equiv=\"refresh\" content=\"" + cfgHTML_REFRESH + "\">\n");
			fw.write("<style type=\"text/css\">\n");
			fw.write("table.clsd2ol { font-family: Arial; }\n");
			fw.write(".clsd2ol th { font-size: 10pt; }\n");
			fw.write(".clsd2ol td { font-size: 10pt; }\n");
			fw.write("td.clsd2olred { background-color: red; }\n");
			fw.write("td.clsd2olorange { background-color: orange; }\n");
			fw.write("</style>\n");
			fw.write("</head>\n");
			fw.write("<body>\n");
			fw.write("FADProgress generated at: " + new Date().toString() + "\n");
			fw.write("<p><table border=1 cellpadding=1 cellspacing=1 class=clsd2ol><tr>");
			for (int i = 0; i < fadtm.getColumnCount(); i++)
			{
				fw.write("<th align=center>" + fadtm.getColumnName(i) + "</th>\n");
			}
			fw.write("</tr>\n");
			for (int i = 0; i < fadtm.getRowCount(); i++)
			{
				fw.write("<tr>");
				for (int k = 0; k < fadtm.getColumnCount(); k++)
				{
					FADCruncher d2c = fadtm.getCruncherAt(i);
					fw.write("  <td align=" + d2c.getColumnHtmlAlignment(k));
/*					if (k == FADCruncher.STATE)
					{
						Color cTemp = d2c.getStateBackground(Color.WHITE);
						if (cTemp == Color.RED)
						{	fw.write(" class=clsd2olred"); }
						else if (cTemp == Color.ORANGE)
						{   fw.write(" class=clsd2olorange"); }
					}
					*/
					fw.write(">" + d2c.getColumnHtml(k) + "</td>\n");
				}
				fw.write("</tr>");
			}
			fw.write("</table>\n");
//			fw.write("<br>Total ready: " + jlPoints.getText());
			fw.write("</body>\n");
			fw.write("</html>\n");
			fw.close();
		} catch (IOException ex) {
			System.err.println("Unable to generate HTML output file." + ex);
		}
	}


	//----------------------------------------------------------------------
	public static void reportError(boolean bFatal, Exception ex, String sMsg)
	{
		Date now = new Date();
		DateFormat formatter = DateFormat.getDateTimeInstance(DateFormat.SHORT,
                                   DateFormat.DEFAULT,
                                   Locale.getDefault());
      System.err.println(formatter.format(now) + ": ERROR, " + sMsg);
      if (ex != null) System.err.println("     " + ex);
      if (bFatal) System.exit(-1);
	}

	// BEGIN INTERNAL CLASSES
	//**********************************************************************
	// Keep "table stuff" in this class and "cruncher data" in FADCruncher
	class FADTableModel extends AbstractTableModel
	{
		public final static int ASCENDING = 1;
		public final static int DESCENDING = -1;

		public int iSortColumn = FADCruncher.RATING;
		public int iSortDirection = DESCENDING;
		// aiWidths, asColNames, and asColTips must match FADCruncher static var order
		public int[] aiWidths = {120, 80, 100, 40, 80, 130, 130, 80, 55, 120}; // tot=935
		private String[] asColNames = {"Name",
				"Job start",
				"Job progress",
				"Hits",
				"Activity",
				"Job",
				"Project",
				"Updated",
				"Rating",
				"Server"};
		private String[] asColTips = {"Node Name",
				"When the job started",
				"Job progress",
				"Hits",
				"Last activity timestamp",
				"Job name",
				"Project name",
				"Progress update received",
				"System rating",
				"Server sending progress"};		
		private java.util.List data = new ArrayList();
		private java.util.List sortIndex = new ArrayList();

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		public int getColumnCount() { return asColNames.length; }

		public String getColumnName(int col) { return asColNames[col]; }

		public String getColumnTip(int col) { return asColTips[col]; }
		
		public Class getColumnClass(int col) { return new FADCruncher("","").getColumnClass(col); }

		public int getRowCount() { return data.size(); }

		public Object getValueAt(int row, int col)
		{
			int iSortedRow = ((Integer) sortIndex.get(row)).intValue();
			return ((FADCruncher)(data.get(iSortedRow))).getColumn(col);
		}

		public Object getUnsortedValueAt(int row, int col)
		{
			return ((FADCruncher)(data.get(row))).getColumn(col);
		}

		public FADCruncher getCruncherAt(int row)
		{
			int iSortedRow = ((Integer) sortIndex.get(row)).intValue();
			return (FADCruncher) data.get(iSortedRow);
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		public void updateClient(String[] asIncoming)
		{
			final int iNOT_FOUND = -1;

			int iCruncherRow = iNOT_FOUND;
			FADCruncher oCruncher = null;
			Object oHoldSortedValue = null; // could be any datatype thus Object class
	//		Object oHoldResultsToUpload = null;

			boolean bNewCruncher = false;

			// look for an existing cruncher that matches the incoming cruncher
			for (int i = 0; i < data.size(); i++)
			{
				oCruncher = (FADCruncher) data.get(i);
				if (oCruncher.sName.equals(asIncoming[PKTNODENAME]) &&
					oCruncher.sServerName.equals(asIncoming[PKTPROGRESSDNAME]))
				{
					iCruncherRow = i; // matched an existing row
					oHoldSortedValue = oCruncher.getColumn(iSortColumn);
//					oHoldResultsToUpload = oCruncher.getColumn(FADCruncher.RESULTSTOUPLOAD);

					break; // done looping
				}
			}

			if (iCruncherRow == iNOT_FOUND) // new cruncher just joined
			{
				bNewCruncher = true; // lets go ahead and refresh the display after adding this
				// create new record and add to list
				oCruncher = new FADCruncher(asIncoming[PKTNODENAME], asIncoming[PKTPROGRESSDNAME]);
				data.add(oCruncher);
				iCruncherRow = data.size() - 1;
				sortIndex.add(new Integer(iCruncherRow)); // to keep data & sortIndex lists consistent, add this immediately
			}

			oCruncher.update(asIncoming);

			if (oCruncher.bNoProgressTxt.booleanValue()) return; // no sense proceeding if all data bad, dont chg existing displayed vals
			if (bNewCruncher) // always sort and update screen when a new cruncher arrives
			{
				sort();
				fireTableDataChanged();
				updatePoints();
			} else {
				// only sort if our sorted value has changed!
				if (oCruncher.compareTo(oHoldSortedValue, iSortColumn) != 0) sort();
				// only loop through all crunchers and sum their resultstoupload if it just chgd
//				if (oCruncher.compareTo(oHoldResultsToUpload, FADCruncher.RESULTSTOUPLOAD) != 0) updatePoints();
			}
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		public void changeSort(int iNewSortColumn)
		{
			if (iSortColumn == iNewSortColumn)
			{
				// toggle ascending, descending
				iSortDirection = (iSortDirection == ASCENDING ? DESCENDING : ASCENDING);
			} else {
				iSortColumn = iNewSortColumn;
				iSortDirection = ASCENDING;
			}
			sort();
			fireTableDataChanged();
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		public void sort()
		{
			// the objective of this fn is to produce a new List into sortIndex
			// reflecting the public vars iSortColumn and iSortDirection.
			// this might have been called because the data changed, or because
			// the user wants to sort differently
			int i,k;
			int iHold;
			int iSize = data.size();
			int iCompared = 0;
			int[] aiSortedIndex = new int[iSize];
			Object[] aoValues = new Object[iSize]; // column data could be any class
			int iSmallestIndex;
			Object oSmallestItem;
			Class clsColumn = getColumnClass(iSortColumn);

			// initialize vals
			for (i = 0; i < iSize; i++)
			{
				aoValues[i] = getUnsortedValueAt(i, iSortColumn);
				aiSortedIndex[i] = i;
			}

			// dont need to sort if not sorting
			if (iSortDirection != 0)
			{
				// do the sort
				for (i = 0; i < iSize - 1; i++)
				{
					// initialize smallest
					iSmallestIndex = i;
					oSmallestItem = aoValues[i];
					// can we find smaller one in remaining items?
					for (k = i + 1; k < iSize; k++)
					{
						// Dont have a smallest item yet?  then take this one
						if (oSmallestItem == null)
						{
							iCompared = 1;
						// Incoming value is null?  then it is not smaller
						} else if (aoValues[k] == null) {
							iCompared = -1;
						// Now use the correct class to compare the item
						} else if (clsColumn == Integer.class) {
							iCompared = ((Integer) oSmallestItem).compareTo((Integer) aoValues[k]);
						} else if (clsColumn == String.class) {
							iCompared = ((String) oSmallestItem).toUpperCase().compareTo(((String) aoValues[k]).toUpperCase());
						} else if (clsColumn == Date.class) {
							iCompared = ((Date) oSmallestItem).compareTo((Date) aoValues[k]);
						} else if (clsColumn == Float.class) {
							iCompared = ((Float) oSmallestItem).compareTo((Float) aoValues[k]);
						}
						// It IS smaller, so replace local (smallest) vars
						if (iCompared > 0)
						{
							iSmallestIndex = k;
							oSmallestItem = aoValues[k];
						}
					}
					// found smaller one?  if so swap 'em
					if (i != iSmallestIndex)
					{
						aoValues[iSmallestIndex] = aoValues[i];
						aoValues[i] = oSmallestItem;
						iHold = aiSortedIndex[iSmallestIndex];
						aiSortedIndex[iSmallestIndex] = aiSortedIndex[i];
						aiSortedIndex[i] = iHold;
					}
				}
			}

			// store in main sortindex
			k = (iSortDirection == ASCENDING ? 0 : iSize - 1);
			for (i = 0; i < iSize; i++)
			{
				sortIndex.set(i, new Integer(aiSortedIndex[k]));
				k += iSortDirection;
			}
		}
	} // end inside class FADTableModel

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	// This class represents one FAD crunching machine, a FADProgressNode
	public class FADCruncher
	{

		// THESE static vars dictate the GUI column order and do not need to match
		// the order of the pkt fields.
		// Table Model's aiWidths, asColNames, and asColTips must match this order
		private String[] asColNames = {"Name",
				"Job start",
				"Job prog",
				"Hits",
				"Activity",
				"Job",
				"Project",
				"Updated",
				"Rating",
				"Server"};
		public final static int NAME = 0;
		public final static int JOBSTART = 1;
		public final static int JOBPROGRESS = 2;
		public final static int HITS = 3;
		public final static int ACTIVITY = 4;
		public final static int JOBNAME = 5;
		public final static int PROJECTNAME = 6;
		public final static int UPDATED = 7;
		public final static int RATING = 8;
		public final static int SERVERNAME = 9;

		// these vars hold this cruncher instances' personal values
		private String sName;
		private Date dtJobStart;
		private Float fJobProgress;
		private Integer iHits;
		private Date dtActivity;
		private String sJobName;
		private String sProjectName;
		private Date dtUpdated;
		private Integer iRating;
		private String sServerName;

		// other vars
		public Boolean bNoProgressTxt; // should we show the no progress being made?
//		public Date dtStateChanged; // datestamp when state last changed - uses local time

		// use this sdf for html output
		private SimpleDateFormat sdfHtml = new SimpleDateFormat("H:mm:ss"); // hopefully all locale-friendly
		// use this sdf for html output when not same day
		private SimpleDateFormat sdfHtmlFull = new SimpleDateFormat("d-MMM H:mm");
		// use this sdf for incoming fad dates
		private SimpleDateFormat sdfFad = new SimpleDateFormat("HH:mm dd-MMM-yy");
		private ArrayList alProjectCodes = new ArrayList();
		private ArrayList alProjectNames = new ArrayList();

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Constructor -- initialize vals
		public FADCruncher(String sNodeName, String sProgressName)
		{
			this.sName = sNodeName;
			this.dtJobStart = new Date();
			this.fJobProgress = new Float(0);
			this.iHits = new Integer(0);
			this.dtActivity = new Date();
			this.sJobName = new String();
			this.sProjectName = new String();
			this.dtUpdated = new Date();
			this.iRating = new Integer(0);
			this.sServerName = sProgressName;
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Called every update to update this cruncher's data values
		public void update(String[] asIncoming)
		{
			dtUpdated = new Date(); // no matter what we got an update!

			// FAD client
			if (!asIncoming[PKTTYPE].equals(sPACKET_TYPE_FAD))
			{
				reportError(false, null, "WRONG PACKET TYPE");
				return;
			}

			bNoProgressTxt = new Boolean(asIncoming[PKTTYPE].equals(sPACKET_TYPE_PROBLEM));
			if (bNoProgressTxt.booleanValue())
			{
				return; // no sense continuing, unlikely any good data
			}
			
			if (!sJobName.equals(asIncoming[PKTFADTARGET].trim()))
			{
				// new job.
				sJobName = asIncoming[PKTFADTARGET].trim(); // assign new jobname
				sProjectName = ""; // clear projectname
				
				if (!findProject())
				{
					retrieveFADProjects(); // get fresh list
					findProject(); // one last try
				}
			}					
			try
			{
				fJobProgress = new Float(Float.parseFloat(asIncoming[PKTFADPROGRESS].replace('%', ' ').trim()) / (float) 100);
			} catch (Exception ex) { fJobProgress = new Float(0f); };
			try
			{
				dtJobStart = sdfFad.parse(asIncoming[PKTFADSTART]);
			} catch (Exception ex) { dtJobStart = null; };
			try
			{
				dtActivity = sdfFad.parse(asIncoming[PKTFADLATEST]);
			} catch (Exception ex) { dtActivity = null; };
			try
			{
				iHits = new Integer(asIncoming[PKTFADHITS].trim());
			} catch (Exception ex) { iHits = new Integer(0); };
			try
			{
				iRating = new Integer(asIncoming[PKTFADRATING].trim());
			} catch (Exception ex) { /* dont bother doing anything since these are only 
				given once every 10 mins */ };
			return;
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// This fn is called to locate our job in the project list
		private boolean findProject()
		{
			int iDash = sJobName.indexOf('-');
			int iPeriod = sJobName.indexOf('.');
			String sTemp = sJobName.substring(iDash + 1, iPeriod);

			int iIndex = alProjectCodes.indexOf(sTemp);
			if (iIndex == -1) return false; // didnt find it

			// did find it
			sProjectName = (String) alProjectNames.get(iIndex);
			return true;
		}
		
		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// This fn is called to format an elapsed time
		private String formatHMS(int iEstimate)
		{
			StringBuffer sb = new StringBuffer("   "); // until I get around to centering in the cell...
			if (iEstimate < 0)
			{
				iEstimate = Math.abs(iEstimate);
				sb.append("-");
			}
			int iHours = iEstimate / 3600;
			int iMinutes = (iEstimate - (iHours * 3600)) / 60;
			int iSeconds = iEstimate - (iHours * 3600) - (iMinutes * 60);

//			if (iHours < 10) sb.append("0");
			sb.append(iHours);
			sb.append(":");
			if (iMinutes < 10) sb.append("0");
			sb.append(iMinutes);
			sb.append(":");
			if (iSeconds < 10) sb.append("0");
			sb.append(iSeconds);
			return sb.toString();
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// This fn is called by the gui to retrieve a particular column's value
		public Object getColumn(int iColumn)
		{
			switch (iColumn)
			{
				case NAME:
					return sName;
				case JOBSTART:
					return dtJobStart;
				case JOBPROGRESS:
					return fJobProgress;
				case HITS:
					return iHits;
				case ACTIVITY:
					return dtActivity;
				case JOBNAME:
					return sJobName;
				case PROJECTNAME:
					return sProjectName;
				case UPDATED:
					return dtUpdated;
				case RATING:
					return iRating;
				case SERVERNAME:
					return sServerName;
				default:
					return null;
			}
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// This fn is called by the html output to retrieve a particular column's value as html
		public String getColumnHtml(int iColumn)
		{
			Object oUnknown = getColumn(iColumn); // grab column value
			if (oUnknown == null) return "&nbsp;"; // jump out if null

			if (iColumn == JOBPROGRESS) // special handling for progress -- needs %
			{
				float f = ((Float) oUnknown).floatValue();
				f *= 100.0;
				return (new Integer((int) f)).toString() + "%";
			}

			String sClassName = getColumnClass(iColumn).getName();
			if (sClassName.equals("java.util.Date")) 
			{
				Date dtValue = (Date) oUnknown;
				Date dtMidnightToday = new Date();
				dtMidnightToday.setHours(0);
				dtMidnightToday.setMinutes(0);
				dtMidnightToday.setSeconds(0);
				if (dtMidnightToday.compareTo(dtValue) > 0)
				{
					// tis not the same day, so show the month and day too
					return sdfHtmlFull.format(dtValue);
				} else {
					// same day, so just show hms
					return sdfHtml.format(dtValue);
				}
			}

			String sTemp = oUnknown.toString();
			if (sTemp.length() == 0) return "&nbsp;";
			return sTemp;
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// This fn is called by the html output to determine column alignment
		public String getColumnHtmlAlignment(int iColumn)
		{
			String sClassName = getColumnClass(iColumn).getName();
			if (sClassName.equals("java.util.Date"))
				return "middle";
			else if (sClassName.equals("java.lang.Integer") ||
						sClassName.equals("java.lang.Float"))
				return "right";
			else
				return "left";
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// This fn is called by the gui to retrieve a particular column's class
		public Class getColumnClass(int iColumn)
		{
			switch (iColumn)
			{
			// date class vars
			case JOBSTART:
			case ACTIVITY:
			case UPDATED:
				return new Date().getClass();
				
			// float class vars
			case JOBPROGRESS:
				return new Float(0f).getClass();

			// int class vars
			case HITS:
			case RATING:
				return new Integer(0).getClass();

			// string class vars
			default:
				return new String().getClass();
			}
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		public int compareTo(Object oValue, int iColumn)
		{
			try // try block to catch null comparison
			{
				switch (iColumn)
				{
					case NAME:
						return sName.compareTo((String) oValue);
					case JOBSTART:
						return dtJobStart.compareTo((Date) oValue);
					case JOBPROGRESS:
						return fJobProgress.compareTo((Float) oValue);
					case HITS:
						return iHits.compareTo((Integer) oValue);
					case ACTIVITY:
						return dtActivity.compareTo((Date) oValue);
					case JOBNAME:
						return sJobName.compareTo((String) oValue);
					case PROJECTNAME:
						return sProjectName.compareTo((String) oValue);
					case UPDATED:
						return dtUpdated.compareTo((Date) oValue);
					case RATING:
						return iRating.compareTo((Integer) oValue);
					case SERVERNAME:
						return sServerName.compareTo((String) oValue);
					default:
						return 0;
				}
			} catch (Exception ex) {
				if (cfgDEBUG) reportError(false, ex, "Handled #compareTo exception.");
				return 0; // dont care about this error
			}
		}

		//----------------------------------------------------------------------
		private void retrieveFADProjects()
		{
			BufferedReader in = null;
			boolean bFoundIt = false;
			
			alProjectCodes = new ArrayList();
			alProjectNames = new ArrayList();

			// plow through latest update to get our records
			try
			{
				URL urlStats = new URL(sSTATS_PAGE);
				URLConnection conn = urlStats.openConnection();
				in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
				String sInLine;
				String sTemp;
				int iIndex;

				while ((sInLine = in.readLine()) != null)
				{
					if (!sInLine.startsWith("<TR><TD><A")) continue; // only care about these lines
					// line looks like this:  <TR><TD><A href="cancer-frame.html">Cancer</A></TD>
					// parse out project name
					iIndex = sInLine.indexOf('>', 10);
					sTemp = sInLine.substring(iIndex + 1);
					iIndex = sTemp.indexOf('<');
					alProjectNames.add(sTemp.substring(0, iIndex));
					
					// now parse next line which has the target code
					// line looks like this:  <TD>1NKP-Q2 </TD>
					sInLine = in.readLine();
					alProjectCodes.add(sInLine.substring(4, 11).toLowerCase());
				} // end while
				in.close();
			} catch (Exception ex) {
				if (in != null) try { in.close(); } catch (Exception ex2) { };
				reportError(false, null, "Problem accessing " + sSTATS_PAGE + " : " + ex);
			}
		}

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Given the default color, override with a new color if necessary due
		// to the state of this cruncher having some kind of issue
/*
		public Color getStateBackground(Color cDefault)
		{
			// show type-problem as red bkgd
			if (bNoProgressTxt.booleanValue()) // problem means red
			{
				return Color.RED;
			// show no chg of state for Xtra long as red bkgd
			} else if (!bIsFad.booleanValue() && dtStateChanged.getTime() + iANCIENT_STATE_DELAY < (new Date()).getTime()) {
				return Color.RED;
				// show no-progress-node-update as orange bkgd
			} else if (dtUpdated.getTime() + iSTALE_UPDATE_DELAY < (new Date()).getTime()) {
				return Color.ORANGE;
			// show no chg of state as orange bkgd
			} else if (!bIsFad.booleanValue() && dtStateChanged.getTime() + iSTALE_STATE_DELAY < (new Date()).getTime()) {
				return Color.ORANGE;
			}
			return cDefault;
		} // end getStateBackgrond 
*/
	} // end FADCruncher

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	public class FADProgressRenderer extends JLabel implements TableCellRenderer
	{
		private float fValue;
		private Color cFill;
		private Color cPercent;
		private NumberFormat pctFormatter;

		public FADProgressRenderer(boolean isBordered)
		{
			super();
			setBorder(BorderFactory.createLoweredBevelBorder());
			cFill = new Color(Integer.parseInt("008B8B", 16));
			cPercent = new Color(Integer.parseInt("000000", 16));
			pctFormatter = NumberFormat.getPercentInstance(Locale.getDefault());
			setOpaque(true);
		}

		public void paintComponent(Graphics g)
		{
			super.paintComponent(g);
			g.setColor(cFill);
			int iFillWidth = (int) (fValue * getWidth());
			g.fillRect(0, 3, iFillWidth, getHeight() - 5);
			String sTemp = pctFormatter.format(fValue);
			g.setColor(cPercent);
			Font fontTemp = g.getFont();
			fontTemp = fontTemp.deriveFont(Font.PLAIN, fontTemp.getSize() - 1);
			g.setFont(fontTemp);
			g.drawChars(sTemp.toCharArray(), 0, sTemp.length(), getWidth() / 3, getHeight() - 4);
		}

		public Component getTableCellRendererComponent(JTable table, Object value,
								boolean selected, boolean focused, int row, int column)
		{
			// alternate row backgrounds
			if (row%2 == 1)
			{
				setBackground(cAlternatingBgnd);
			} else {
				setBackground(table.getBackground());
			}
			fValue = ((Float) value).floatValue();
			return this;
		}
	} // end FADProgressRenderer

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	public class FADTextRenderer extends JLabel implements TableCellRenderer
	{
		public FADTextRenderer()
		{
			super();
			setHorizontalAlignment(javax.swing.SwingConstants.LEFT);
			setFont(fontRegular);
			setOpaque(true);
		}

		public Component getTableCellRendererComponent(JTable table, Object value,
								boolean selected, boolean focused, int row, int column)
		{
			if (row%2 == 1)
			{
				setBackground(cAlternatingBgnd);
			} else {
				setBackground(table.getBackground());
			}
			if (value == null)
				setText("");
			else
				setText(value.toString());
			return this;
		}
	} // end FADTextRenderer

	//----------------------------------------------------------------------
	public class FADStateRenderer extends FADTextRenderer implements TableCellRenderer
	{
		public FADStateRenderer()
		{
			super();
		}

		public Component getTableCellRendererComponent(JTable table, Object value,
								boolean selected, boolean focused, int row, int column)
		{
			// all rows get an alternating background color
			if (row%2 == 1)
			{
				setBackground(cAlternatingBgnd);
			} else {
				setBackground(table.getBackground());
			}
			String sValue = (String) value;
			FADCruncher dfc = ((FADTableModel) table.getModel()).getCruncherAt(row);
			// change background color if trouble with this cruncher
//			setBackground(dfc.getStateBackground(this.getBackground()));

			setText(value.toString());
			return this;
		}

	} // end FADStateRenderer

	//----------------------------------------------------------------------
	public class FADDateRenderer extends JLabel implements TableCellRenderer
	{
		private DateFormat df;
		private SimpleDateFormat sdf;
		private SimpleDateFormat sdfFull;
		private boolean bShowStale;

		public FADDateRenderer()
		{
			this(false);
		}

		public FADDateRenderer(boolean bShowStale)
		{
			super();
			this.bShowStale = bShowStale;
			setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
			setFont(fontRegular);
			df = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.MEDIUM, Locale.getDefault());
			sdf = new SimpleDateFormat("H:mm:ss"); // hopefully all locale-friendly
			sdfFull = new SimpleDateFormat("d-MMM H:mm");
			setOpaque(true);
		}

		public Component getTableCellRendererComponent(JTable table, Object value,
								boolean selected, boolean focused, int row, int column)
		{
			// all rows get an alternating background color
			if (row%2 == 1)
			{
				setBackground(cAlternatingBgnd);
			} else {
				setBackground(table.getBackground());
			}

			// if this instance was setup to show stale dates, then handle here
			if (bShowStale)
			{
				FADCruncher dfc = ((FADTableModel) table.getModel()).getCruncherAt(row);
				if (dfc.bNoProgressTxt.booleanValue())
				{
					setBackground(Color.RED);
				} else if (value != null) {
					if (((Date) value).getTime() + iSTALE_UPDATE_DELAY < (new Date()).getTime())
						setBackground(Color.ORANGE);
				}
			}

			if (value == null)
			{
				setText("");
			} else {
				Date dtValue = (Date) value;
				Date dtMidnightToday = new Date();
				dtMidnightToday.setHours(0);
				dtMidnightToday.setMinutes(0);
				dtMidnightToday.setSeconds(0);
				if (dtMidnightToday.compareTo(dtValue) > 0)
				{
					// tis not the same day, so show the month and day too
					setText(sdfFull.format(dtValue));
				} else {
					// same day, so just show hms
					setText(sdf.format(dtValue));
				}
			}
			return this;
		}
	} // end FADDateRenderer

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	public class FADNumberRenderer extends JLabel implements TableCellRenderer
	{
		private boolean bIsFloat = false;

		public FADNumberRenderer()
		{
			this(false);
		}
		public FADNumberRenderer(boolean bIsFloat)
		{
			this.bIsFloat = bIsFloat;
			setHorizontalAlignment(javax.swing.SwingConstants.RIGHT);
			setFont(fontRegular);
			setOpaque(true);
		}


		public Component getTableCellRendererComponent(JTable table, Object value,
								boolean selected, boolean focused, int row, int column)
		{
			Object vValue;

			if (row%2 == 1)
			{
				setBackground(cAlternatingBgnd);
			} else {
				setBackground(table.getBackground());
			}

			// display the value
			if (value == null)
			{
				setText("");
			} else if (bIsFloat) {
				vValue = ((Float) value);
				setText(nFormat.format(vValue));
			} else {
				vValue = ((Integer) value);
				setText(nFormat.format(vValue));
			}

			return this;
		}
	} // end FADNumberRenderer

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	public class FADHeaderRenderer extends JLabel implements TableCellRenderer
	{
		private FADArrow arrowUp;
		private FADArrow arrowDown;

		public FADHeaderRenderer()
		{
			super();
			setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
			setHorizontalTextPosition(LEADING);
			setBorder(UIManager.getBorder("TableHeader.cellBorder"));
			setFont(fontRegular);
			setIconTextGap(2);
			arrowUp = new FADArrow(true);
			arrowDown = new FADArrow(false);
			setOpaque(true);
		}

		public Component getTableCellRendererComponent(JTable table, Object value,
								boolean selected, boolean focused, int row, int column)
		{
			setText((value == null) ? "" : value.toString());
			int iColumnOfModel = table.convertColumnIndexToModel(column);
			setIcon(iColumnOfModel == fadtm.iSortColumn ?
				(fadtm.iSortDirection == FADTableModel.ASCENDING ? arrowUp : arrowDown) :
				null);
			return this;
		}
	} // end FADHeaderRenderer

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	public class FADArrow implements Icon
	{
		private Color cSortIcon;
		private Polygon pUp;
		private Polygon pDown;
		private boolean bUp;

		public FADArrow(boolean bUp)
		{
			this.bUp = bUp;
			cSortIcon = new Color(Integer.parseInt("008B8B", 16));
			int[] xPointsUp = {0, 8, 4, 0};
			int[] yPointsUp = {11, 11, 3, 11};
			pUp = new Polygon(xPointsUp, yPointsUp, xPointsUp.length);
			int[] xPointsDown = {0, 8, 4, 0};
			int[] yPointsDown = {3, 3, 11, 3};
			pDown = new Polygon(xPointsDown, yPointsDown, xPointsDown.length);
		}

		public void paintIcon(Component c, Graphics g, int x, int y)
		{
			g.translate(x, y);
			g.setColor(cSortIcon);
			g.drawPolygon(bUp ? pUp : pDown);
			g.translate(-x, -y);
		}
		public int getIconWidth() { return 12; }
		public int getIconHeight() { return 12; }
	}

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	private class FADHeaderMouseHandler extends MouseAdapter
	{
		public void mouseClicked(MouseEvent e)
		{
			JTableHeader jheader = (JTableHeader) e.getSource();
			TableColumnModel columnModel = jheader.getColumnModel();
			int iViewColumn = columnModel.getColumnIndexAtX(e.getX());
			int iColumn = columnModel.getColumn(iViewColumn).getModelIndex();
			if (iColumn != -1)
			{
				fadtm.changeSort(iColumn);
			}
		}
	} // end FADHeaderMouseHandler

	//----------------------------------------------------------------------
	//----------------------------------------------------------------------
	// This is the class that runs in another thread.  It sits in a loop waiting
	// for datagram packets to arrive, then calls code to handle the packet.
	public class FADListenThread extends Thread
	{
		protected DatagramSocket socket = null;

		public FADListenThread(String sName) throws IOException
		{
			super(sName);
			try
			{
				socket = new DatagramSocket(cfgPORT);
			} catch (BindException ex) {
				reportError(true, ex, "Unable to setup socket on port " + cfgPORT + "; possibly in use already.");
			}
		}

		public void run()
		{
			while (true)
			{
				try
				{
					byte[] buff = new byte[512];
	// TODO:  check for buffer received too big

					// wait, get packet from crunch machine
					DatagramPacket packet = new DatagramPacket(buff, buff.length);
					socket.receive(packet); // this will wait for a packet

					String sIncoming = new String(packet.getData());
					if (cfgDEBUG) System.out.println(sIncoming);
					String[] asIncoming = sIncoming.split(sPACKET_DELIM);
					// if packet not correct version spit out msg and ignore it
					if (!asIncoming[PKTVERSION].equals(sPACKET_VERSION))
					{
						reportError(false, null, "invalid FAD packet version.  Obsolete FADProgressNode(?) sent this: \n" +
							sIncoming +
							"\nFROM: " + packet.getAddress());
					// else packet has correct version
 					} else {
						fadtm.updateClient(asIncoming);
					}
				} catch (IOException ex) {
					reportError(true, ex, "error accepting datagram packets");
				}
			} // end while
		} // end run
	} // end FADListenThread

} // end class FADProgress

